
/*
 * stargate.c
 * Stargate version 0.15 for VGA Planets
 * This program is Copyright (c) 1999-2003 Ilkka Virta and others
 * Distribution allowed under the terms of the
 * GNU General Public License version 2 (included in file copying)
 *
 * Uses PHost Development Kit (PDK).
 * Has been known to compile and work with DJGPP on a Windows NT 4 box and
 * is supposed to work on some kind of Unixen.
 *
 */

#include<stdio.h>
#include<stdlib.h>
#include<errno.h>
#include"phostpdk.h"
#include<string.h>
#include<math.h>
#include<time.h>

#define min(x,y) (((x) < (y)) ? (x) : (y))
#define max(x,y) (((x) > (y)) ? (x) : (y))

#define PACKED __attribute__ ((packed))

#ifdef __sun__
#if 0
extern char *sys_errlist[];
#endif
#endif

struct s_stargate_config {
        int buildcost1[NumCargoTypes]; /* First stage cost - cannot recycle */
        int buildcost2[NumCargoTypes]; /* Second stage cost - recyclable */
                                       /* Total cost is buildcost1 + buildcost2 */
        int mindistfromplanet;
        int minhulltechtobuild;
        int minenginetechtouse;
        int minbeamtechtodestroy;
        int maxscandist;
        int maxseedist;
        int mintravel;
        int maxtravel;
        int usefuelburn;
        int lost_in_hyperspace;
        int alliancesupport;
        int tellsupport;
        int towsupport;
        int allowwarpzerotow;

        int allow_buildsg_p[RACE_NR+1];
        int allow_buildsg_h[HULL_NR+1];
        int allow_deploysg_p[RACE_NR+1];
        int allow_deploysg_h[RACE_NR+1];
        int allow_usesg_p[RACE_NR+1];
        int allow_usesg_h[HULL_NR+1];
        int allow_destroysg_p[RACE_NR+1];
        int allow_destroysg_h[HULL_NR+1];
        int verbosity;
        int misn_buildsg;
        int misn_deploysg;
        int misn_usesg;
        int misn_destroysg;

        int restrict_interracial_travelling;
};

union s_racebits {
        Uns16 w PACKED;
#if 0
        struct {
                Uns16           :1;
                Uns16 federal   :1;
                Uns16 lizard    :1;
                Uns16 romulan   :1;
                Uns16 klingon   :1;
                Uns16 privateer :1;
                Uns16 borg      :1;
                Uns16 tholian   :1;
                Uns16 empire    :1;
                Uns16 robotic   :1;
                Uns16 rebel     :1;
                Uns16 colonial  :1;
                Uns16           :4;
        } b PACKED;
#endif
};

struct s_stargatehst_header {
        char stargates[10] PACKED;
        Uns32 version PACKED;
        Uns32 headerlen PACKED;
        Uns32 footerlen PACKED;
        Uns16 lastid PACKED;
        Uns32 numgates PACKED;
        Uns32 ptr2gates PACKED;
        Uns32 numships PACKED;
        Uns32 ptr2ships PACKED;
} PACKED;

struct s_stargate {
        Uns16 id PACKED;
        Uns16 x,y PACKED;
        Uns16 builtby PACKED;
        union s_racebits knownby PACKED;
} PACKED;

struct s_stargatedata {
        int numgates;
        struct s_stargate *gates;
        int numships;
        Uns16 *ships;
        int lastid;
};

Uns16 ShipFreeCargo(Uns16 shipid);
int read_config(const char *configfilename, struct s_stargate_config *config);
int read_sgdata(struct s_stargatedata *stargatedata);
int destroystargate(int shipid, struct s_stargatedata *data,
                    struct s_stargate_config *config);
int usestargate(int shipid, struct s_stargatedata *data,
                    struct s_stargate_config *config);
int checkshipcargo(int shipid, struct s_stargatedata *data,
                   struct s_stargate_config *config);
int deploystargate(int shipid, struct s_stargatedata *data,
                    struct s_stargate_config *config);
int buildstargate(int shipid, struct s_stargatedata *data,
                    struct s_stargate_config *config);
int scanstargates(int shipid, struct s_stargatedata *data,
                    struct s_stargate_config *config);
int sendgatemessages(struct s_stargatedata *data,
                    struct s_stargate_config *config);
int alliedvision(struct s_stargatedata *data,
                 struct s_stargate_config *config);
int tellcommands(struct s_stargatedata *data,
                 struct s_stargate_config *config);
int sendconfigmessages(struct s_stargate_config *config);
struct s_stargate *low_addgate2list(int listlength,
                                    struct s_stargate *gates);
Uns16 *low_addship2list(int listlength, Uns16 *ships);
int low_cleangatelist(int num, struct s_stargate *gates);
int low_cleanshiplist(int num, Uns16 *ships);
int rtrim(char *str, const char *chars);
int write_sgdata(struct s_stargatedata *stargatedata);

const Uns32 program_version = 0x00000015;
const char program_name[] = "Stargate version 0.15";
const char copyright[] = "Copyright (c) 1999-2003 Ilkka Virta GNU GPL";

FILE *logfile;

int main(int argc, char *argv[])
{
        int i;
        struct s_stargate_config config;
        struct s_stargatedata stargatedata;
        char *configfilename;
        int phase = 0; /* 1 = auxhost 1 or with pcontrol just before movement */
                       /* 2 = auxhost 2 */
        char *logfilename;
        time_t today;
        const char *mode;

        fprintf(stdout, "\n%s %s\n", program_name, copyright);


        if (argc > 5 || argc < 3) {
                fprintf(stderr, "Usage: %s <phase> <gamedirectory>\n"
                                "       [<rootdirectory> [<configfilename]]\n", argv[0]);
                exit(EXIT_FAILURE);
        }

        if (argc == 4) {
                gRootDirectory = argv[3];
                gGameDirectory = argv[2];
                phase = atoi(argv[1]);
        } else if (argc == 3) {
                gGameDirectory = argv[2];
                phase = atoi(argv[1]);
        }
        if (argc == 5) {
                configfilename = argv[4];
                gRootDirectory = argv[3];
                gGameDirectory = argv[2];
                phase = atoi(argv[1]);
        } else {
                if ((configfilename = malloc(strlen(gGameDirectory) + 20)) == NULL) {
                        Error("Cannot allocate memory: %s\n", strerror(errno));
                        exit(EXIT_FAILURE);
                }
#ifdef MSDOS
                sprintf(configfilename, "%s\\stargate.cfg", gGameDirectory);
#else
                sprintf(configfilename, "%s/stargate.cfg", gGameDirectory);
#endif
        }

        if ((logfilename = malloc(strlen(gGameDirectory) + 20)) == NULL) {
                Error("Cannot allocate memory: %s\n", strerror(errno));
                exit(EXIT_FAILURE);
        }

        today = time(NULL);
        mode = (phase == 1) ? "w" : "a";
#ifdef MSDOS
        sprintf(logfilename, "%s\\stargate.log", gGameDirectory);
#else
        sprintf(logfilename, "%s/stargate.log", gGameDirectory);
#endif
        if ((logfile = fopen(logfilename, mode)) == NULL) {
                if ((logfile = fopen(logfilename, "w")) == NULL) {
                        Error("Cannot open logfile (stargate.log): %s\n",
                          strerror(errno));
                        exit(EXIT_FAILURE);
                }
        }

        gLogFile = logfile;

        /* Initialize the PDK */
        InitPHOSTLib();
        InitializeShips();
        if ((! ReadGlobalData()) || (! ReadHostData())) {
                fprintf(logfile, "Reading host data files failed.\n");
                exit(EXIT_FAILURE);
        }

        fprintf(logfile, "\n---\n%s executed on %s\n"
                         "  turn %d\n"
                         "  gamedir %s\n"
                         "  rootdir %s\n"
                         "  phase %d\n", program_name, ctime(&today),
                                         TurnNumber(), gGameDirectory,
                                         gRootDirectory, phase);

        fprintf(stdout, "\n---\n%s executed on %s\n"
                         "  turn %d\n"
                         "  gamedir %s\n"
                         "  rootdir %s\n"
                         "  phase %d\n", program_name, ctime(&today),
                                         TurnNumber(), gGameDirectory,
                                         gRootDirectory, phase);

        srand(time(NULL));

        fprintf(logfile, "\nReading config... ");
        read_config(configfilename, &config);
        fprintf(logfile, "ok\nReading stargate file... ");
        read_sgdata(&stargatedata);

        fprintf(logfile, "ok\nProcessing phase %d\n", phase);
        if (phase == 1) {
                fprintf(logfile, "Sending config messages... ");
                sendconfigmessages(&config);
                fprintf(logfile, "ok\nDestroying stargates... ");
                for ( i = 1 ; i <= SHIP_NR ; i++ ) {
                        if (! IsShipExist(i))
                                continue;
                        if (ShipMission(i) == config.misn_destroysg)
                                destroystargate(i, &stargatedata, &config);
                }
                fprintf(logfile, "ok\nUsing stargates... ");
                for ( i = 1 ; i <= SHIP_NR ; i++ ) {
                        if (! IsShipExist(i))
                                continue;
                        if (ShipMission(i) == config.misn_usesg)
                                usestargate(i, &stargatedata, &config);
                }
                fprintf(logfile, "ok\nChecking ship cargoes and deploying stargates... ");
                for ( i = 1 ; i <= SHIP_NR ; i++ ) {
                        if (! IsShipExist(i))
                                continue;
                        checkshipcargo(i, &stargatedata, &config);
                        if (ShipMission(i) == config.misn_deploysg)
                                deploystargate(i, &stargatedata, &config);
                }
                fprintf(logfile, "ok\nBuilding stargates... ");
                for ( i = 1 ; i <= SHIP_NR ; i++ ) {
                        if (! IsShipExist(i))
                                continue;
                        if (ShipMission(i) == config.misn_buildsg)
                                buildstargate(i, &stargatedata, &config);
                }
                fprintf(logfile, "ok\nEnding phase 1\n");
        } else if (phase == 2) {
                fprintf(logfile, "Checking config for alliance support... ");
                if (config.alliancesupport) {
                        fprintf(logfile, "enabled\nSharing information for allies... ");
                        alliedvision(&stargatedata, &config);
                        fprintf(logfile, "ok\n");
                } else
                        fprintf(logfile, "disabled\n");
                fprintf(logfile, "Checking config for tell support... ");
                if (config.tellsupport) {
                        fprintf(logfile, "enabled\nDoing some telling... ");
                        if (tellcommands(&stargatedata, &config))
                                fprintf(stderr, "Cannot process Tell -commands (xterncmd.ext missing)\n");
                        fprintf(logfile, "ok\n");
                } else
                        fprintf(logfile, "disabled\n");

                fprintf(logfile, "Sending messages for known gates... ");
                sendgatemessages(&stargatedata, &config);
                fprintf(logfile, "ok\nChecking ship cargoes and scanning for gates... ");
                for ( i = 1 ; i <= SHIP_NR ; i++ ) {
                        if (! IsShipExist(i))
                                continue;
                        checkshipcargo(i, &stargatedata, &config);
                        scanstargates(i, &stargatedata, &config);
                }
                fprintf(logfile, "ok\nEnding phase 2\n");
        }

        fprintf(logfile, "Writing stargate data back to file... ");
        write_sgdata(&stargatedata);
        fprintf(logfile, "ok\n");
        free(stargatedata.gates);
        free(stargatedata.ships);
        stargatedata.gates = NULL;
        stargatedata.ships = NULL;
        stargatedata.numships = stargatedata.numships = 0;


        fprintf(logfile, "Writing host data back and freeing PHostLib... ");
        WriteHostData();
        FreePHOSTLib();

        fprintf(logfile, "ok\nExiting.\n\n");
        fprintf(stdout, "Exiting.\n\n");

        return 0;

}

int tellcommands(struct s_stargatedata *data,
                 struct s_stargate_config *config)
{
        FILE *f;
        char buf[80];
        int gate, teller, tellee;
        int index;
        int i;

        char *message = NULL;

        sprintf(buf, "%s/xterncmd.ext", gGameDirectory);
        if ((f = fopen(buf, "r")) == NULL)
                return 1;

        if (( message = malloc(600)) == NULL) {
                fprintf(stderr, "Memory allocation failed: %s\n", strerror(errno));
                exit(EXIT_FAILURE);
        }

        fgets(buf, 80, f);
        while (1) {
                fgets(buf, 80, f);
                if (feof(f))
                        break;
                if (sscanf(buf, "%d: stargate tell %d about %d", &teller, &tellee, &gate) != 3)
/*                        break; */
                        continue;
                for ( i = 0 ; i < data->numgates ; i++ )
                        if (data->gates[i].id == gate) {
                                index = i;
                                goto foundgate;
                        }
                continue;
            foundgate:
                if (! (data->gates[index].knownby.w & (0x01 << teller)))
                        continue;
                data->gates[index].knownby.w |= (0x01 << tellee);
                sprintf(message,
"(-s0000)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"We have told %s\x0d"
"about stargate ID#%d\x0d",
                  RaceShortName(tellee, buf), gate);
                WriteAUXHOSTMessage(teller, message);
        }
        free(message);
        fclose(f);
        return 0;
}

int destroystargate(int shipid, struct s_stargatedata *data,
                    struct s_stargate_config *config)
{
        int i, j;
        char *message = NULL;


        if (IsShipExist(shipid)) {
                if ((message = malloc(600)) == NULL) {
                        fprintf(stderr, "Memory allocation failed: %s\n", strerror(errno));
                        exit(EXIT_FAILURE);
                }

                if (config->allow_destroysg_p[ShipOwner(shipid)] == 0 ) {
                            sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Game configuration says thou\x0d"
"may not destroy stargates,\x0d"
"master.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        PutShipMission(shipid, 5);
                        free(message);
                        return 1;
                }

                if (config->allow_destroysg_h[ShipHull(shipid)] == 0 ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Game configuration says this\x0d"
"type of ship can not destroy\x0d"
"stargates, master.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        PutShipMission(shipid, 5);
                        free(message);
                        return 1;
                }


                if (ShipBeamNumber(shipid) == 0 ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We cannot destroy any stargates,\x0d"
"BECAUSE WE HAVE NO WEAPONS!\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        PutShipMission(shipid, 5);
                        free(message);
                        return 1;
                }

                if (BeamTechLevel(ShipBeamType(shipid)) < config->minbeamtechtodestroy ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Our beams' tech level (%u) is too\x0d"
"low to destroy stargates\x0d"
"Tech level %u is the minimum.\x0d",
                          shipid, shipid, ShipName(shipid, NULL),
                          BeamTechLevel(ShipBeamType(shipid)),
                          config->minbeamtechtodestroy);
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        PutShipMission(shipid, 5);
                        free(message);
                        return 1;
                }


                if (ShipCargo(shipid, NEUTRONIUM) == 0) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We do not enough fuel to\x0d"
"destroy any stargates.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        PutShipMission(shipid, 5);
                        free(message);
                        return 1;
                }



                for ( i = 0 ; i < data->numgates ; i++ ) {
                        if (ShipLocationX(shipid) == data->gates[i].x &&
                            ShipLocationY(shipid) == data->gates[i].y ) {
                                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We have destroyed stargate ID#%u\x0d"
"at (%04u, %04u)\x0d",
                  shipid, shipid, ShipName(shipid, NULL), data->gates[i].id,
                  data->gates[i].x, data->gates[i].y);
                                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                                PutShipMission(shipid, 5);

                                for ( j = 1 ; j <= RACE_NR ; j++ ) {
                                        if ( j == ShipOwner(shipid) )
                                                continue;
                                        sprintf(message,
"(-s0000)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"Stargate explosion detected\x0d"
"at (%u, %u)!\x0d", data->gates[i].x, data->gates[i].y);
                                        WriteAUXHOSTMessage(j, message);
                                }
                                data->gates[i].x = data->gates[i].y = 0;
                        }
                }
                free(message);
        }
        return 0;
}

int playerisfriendlyto(int a, int b, int c)
{
        if (a == b)
                return 1;
        if (PlayersAreAllies(a, b) &&
            PlayerAllowsAlly(a, b, c))
                return 1;
        return 0;
}


int usestargate(int shipid, struct s_stargatedata *data,
                    struct s_stargate_config *config)
{
        int i;
        int mindistsquared = config->mintravel * config->mintravel;
        int maxdistsquared = config->maxtravel * config->maxtravel;
        int distancesquared;
        int sourceindex;
        int targetindex;
        int sourceid;
        int targetid;
        int toweeid;
        char *message = NULL;
        int gasneeded = 0;
        Uns16 shiplocationx;
        Uns16 shiplocationy;


        if (IsShipExist(shipid)) {
                shiplocationx = ShipLocationX(shipid);
                shiplocationy = ShipLocationY(shipid);

                toweeid = ShipTowTarget(shipid);

                gasneeded = config->usefuelburn;

                if (( message = malloc(600)) == NULL) {
                        fprintf(stderr, "Memory allocation failed: %s\n", strerror(errno));
                        exit(EXIT_FAILURE);
                }

                PutShipMission(shipid, 5);

                if (toweeid > 0 && toweeid != shipid && IsShipExist(toweeid)) {
                        gasneeded *= 2;
                }

                if (config->allow_usesg_p[ShipOwner(shipid)] == 0) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Game configuration says that\x0d"
"you may not use stargates.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }

                if (config->allow_usesg_h[ShipHull(shipid)] == 0 ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Game configuration says this\x0d"
"type of ship can't use\x0d"
"stargates.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }

                if (toweeid > 0 && toweeid != shipid && IsShipExist(toweeid)) {
                        if ( ! config->towsupport) {
                                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Towing is disabled in game settings,\x0d"
"ignoring jump command.\x0d",
                                  shipid, shipid, ShipName(shipid, NULL));
                                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                                free(message);
                                return 1;
                        }
                }

                if (toweeid > 0 && toweeid != shipid && IsShipExist(toweeid) && ! gPconfigInfo->OneTow &&
                    HullEngineNumber(ShipHull(shipid)) == 1 ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Cannot tow with only one engine\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }

                if (EngTechLevel(ShipEngine(shipid)) < config->minenginetechtouse) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Our engines are too weak to\x0d"
"open any stargates.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }


                if (toweeid > 0 && toweeid != shipid) {
                        if ( ! (IsShipExist(toweeid) &&
                               ShipLocationX(toweeid) == ShipLocationX(shipid) &&
                               ShipLocationY(toweeid) == ShipLocationY(shipid))) {
                                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"The ship targeted to tow is not\x0d"
"present at our location.\x0d"
"Disengaging jump.\x0d",
                                  shipid, shipid, ShipName(shipid, NULL));
                                WriteAUXHOSTMessage(ShipOwner(shipid), message);

                                free(message);
                                return 1;
                        }
                }


                if (ShipCargo(shipid, NEUTRONIUM) <= gasneeded) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We do not have enough fuel to\x0d"
"complete our stargate travel.\x0d"
"(we have    %d kt,\x0d"
" would need %d kt (+ 1 kt))\x0d",
                          shipid, shipid, ShipName(shipid, NULL),
                          ShipCargo(shipid, NEUTRONIUM), gasneeded);
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }

                for ( i = 0 ; i < data->numgates ; i++ ) {
                        if (ShipLocationX(shipid) == data->gates[i].x &&
                            ShipLocationY(shipid) == data->gates[i].y ) {
                                sourceindex = i;
                                sourceid = data->gates[i].id;
                                goto foundfirstgate;
                        }
                }

                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"There is no stargate at\x0d"
"our current position.\x0d"
"(%04d,%04d)",
                  shipid, shipid, ShipName(shipid, NULL),
                  ShipLocationX(shipid), ShipLocationY(shipid));
                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                free(message);
                return 1;


/* Ok, the ship is positioned over a stargate, now see if we can use it. */
        foundfirstgate:

                for ( i = 0 ; i < data->numgates ; i++ ) {
                        if (data->gates[i].id == ShipInterceptTarget(shipid) &&
                            data->gates[i].knownby.w & (0x01 << ShipOwner(shipid))) {
                                targetindex = i;
                                targetid = data->gates[i].id;
                                goto foundsecondgate;
                        }
                }

                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"The target stargate is not\x0d"
"known to us. We cannot proceed.\x0d",
                  shipid, shipid, ShipName(shipid, NULL));
                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                free(message);
                return 1;



/* Target gate is also found, and known, let's see how far it is */
        foundsecondgate:

                distancesquared = DistanceSquared(
                  ShipLocationX(shipid), ShipLocationY(shipid),
                  data->gates[targetindex].x, data->gates[targetindex].y);

                if (distancesquared < mindistsquared ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"The target stargate is too\x0d"
"near to our current position.\x0d"
"(distance is %.1f ly,\x0d"
" minimum is %d.0 ly)\x0d",
                          shipid, shipid, ShipName(shipid, NULL),
                          sqrt(distancesquared), config->mintravel);
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }

                if (distancesquared > maxdistsquared ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"The target stargate is too\x0d"
"far from our current position.\x0d"
"(distance is %.1f ly,\x0d"
" maximum is %d.0 ly)\x0d",
                          shipid, shipid, ShipName(shipid, NULL),
                          sqrt(distancesquared), config->maxtravel);
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }

                /* check for travelling between gates of non-allied players */
                /* Idea from Heiko Schlenker */
                if (config->restrict_interracial_travelling) {
                        int sourcebuilder, targetbuilder, shipowner;
                        sourcebuilder = data->gates[sourceindex].builtby;
                        targetbuilder = data->gates[targetindex].builtby;
                        shipowner = ShipOwner(shipid);

                        if (playerisfriendlyto(sourcebuilder, shipowner, ALLY_SHIPS) &&
                            playerisfriendlyto(targetbuilder, shipowner, ALLY_SHIPS) &&
                            playerisfriendlyto(sourcebuilder, targetbuilder, ALLY_SHIPS) &&
                            playerisfriendlyto(targetbuilder, sourcebuilder, ALLY_SHIPS)) {
                                        /* players are allies. everything is ok. */
                        } else {
                                        /* players are not allies */
                                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"To use a stargate, the owners of the\x0d"
"source and target stargate have to\x0d"
"offer a ship level alliance to each\x0d"
"other. The owner of the travelling\x0d"
"ship and the owners of the stargates\x0d"
"must be allies and must have the ship\x0d"
"level alliance enabled to each other.\x0d"
"We cannot proceed.\x0d",
                                        shipid, shipid, ShipName(shipid, NULL));
                                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                                free(message);
                                return 1;
                        }
                }

/* Ok, ship can move through the gate. If the target still exists... */

                if (data->gates[targetindex].x == 0) {
                        if (config->lost_in_hyperspace) {
                                /* Blow the ship to hell */
                                shiplocationx = ShipLocationX(shipid);
                                shiplocationy = ShipLocationY(shipid);
                                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"%s ID#%u\x0d"                                /* sender's name and id */
"has not been seen after it\x0d"
"entered a stargate at\x0d"
"(%u, %u).\x0d\x0d"                           /* coordinates */
"It is improbable that it\x0d"
"will be seen ever again.\x0d\x0d"
"Possible reason: The ship\x0d"
"couldn't find the target stargate\x0d"
"(because it was destroyed).\x0d",
                                  shipid, ShipName(shipid, NULL), shipid,
                                  shiplocationx, shiplocationy);
                                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                                DeleteShip(shipid);


                                if (toweeid > 0 && toweeid != shipid && IsShipExist(toweeid)) {
                                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"%s ID#%u\x0d"                                /* sender's name and id */
"has not been seen after it\x0d"
"was towed to a stargate at\x0d"
"(%u, %u).\x0d\x0d"                           /* coordinates */
"It is improbable that it\x0d"
"will be seen ever again.\x0d\x0d"
"Possible reason: The tower\x0d"
"couldn't find the target stargate\x0d"
"(because it was destroyed).\x0d",
                                          toweeid, ShipName(toweeid, NULL), toweeid,
                                          shiplocationx, shiplocationy);
                                        WriteAUXHOSTMessage(ShipOwner(toweeid), message);
                                        DeleteShip(toweeid);
                                }

                                free(message);
                                return 1;
                        } else {
                                /* Don't blow the ship, just inform of the
                                   destroyed gate */
                                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's name and id */
"The target stargate is\x0d"
"destroyed. We cannot proceed\x0d"
"with the jump.\x0d",

                                  shipid, shipid, ShipName(shipid, NULL));
                                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                                free(message);
                                return 1;
                        }
                }



                if (toweeid > 0 && toweeid != shipid && IsShipExist(toweeid)) {
                        if ( ! (ShipOwner(shipid) == ShipOwner(toweeid) ||
                                ShipCargo(toweeid, NEUTRONIUM) == 0     ||
                                (config->allowwarpzerotow && ! ShipSpeed(toweeid)) ||
        (PlayersAreAllies(ShipOwner(shipid), ShipOwner(toweeid)) &&
         PlayerAllowsAlly(ShipOwner(toweeid), ShipOwner(shipid), ALLY_SHIPS))
                           )) {
                                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Our towee does not want to be towed.\x0d",
                                  shipid, shipid, ShipName(shipid, NULL));
                                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                                free(message);
                                return 1;
                        }
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's name and id */
"We have successfully towed\x0d"
"ship ID#%u through\x0d"
"stargate ID#%u at (%u, %u)\x0d"
"to stargate ID%u at (%u, %u).\x0d"
"The operation burned %dkt\x0d"
"of neutronium.\x0d",

                          shipid, shipid, ShipName(shipid, NULL), toweeid, sourceid,
                          ShipLocationX(shipid), ShipLocationY(shipid),
                          data->gates[targetindex].id,
                          data->gates[targetindex].x, data->gates[targetindex].y,
                          2 * config->usefuelburn);
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);

                        if (ShipOwner(shipid) != ShipOwner(toweeid)) {
                                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's name and id */
"We have been towed by ship\x0d"
"ID#%u through stargate\x0d"
"ID#%u at (%u, %u) to stargate\x0d"
"ID#%u at (%u, %u).\x0d",
                          toweeid, toweeid, ShipName(toweeid, NULL), shipid,
                          sourceid, ShipLocationX(shipid), ShipLocationY(shipid),
                          data->gates[targetindex].id,
                          data->gates[targetindex].x, data->gates[targetindex].y);
                                WriteAUXHOSTMessage(ShipOwner(toweeid), message);
                        }

                        PutShipMission(shipid, 7);
                        PutShipLocationX(shipid, data->gates[targetindex].x);
                        PutShipLocationY(shipid, data->gates[targetindex].y);
                        PutShipLocationX(toweeid, data->gates[targetindex].x);
                        PutShipLocationY(toweeid, data->gates[targetindex].y);
                        PutShipSpeed(shipid, 0);
                        PutShipSpeed(toweeid, 0);
                        PutShipRelWaypointX(shipid, 0);
                        PutShipRelWaypointY(shipid, 0);
                        PutShipRelWaypointX(toweeid, 0);
                        PutShipRelWaypointY(toweeid, 0);
                        PutShipCargo(shipid, NEUTRONIUM, ShipCargo(shipid, NEUTRONIUM) -
                          gasneeded);

                        free(message);
                        return 0;
                }


/* Now doing the jump. There were 10 things that could have gone wrong */
/* but they didn't. */



                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's name and id */
"We have successfully travelled\x0d"
"from stargate ID#%u\x0d"
"at (%u, %u) to stargate ID%u\x0d"
"at (%u, %u).\x0d"
"The operation burned %dkt\x0d"
"of neutronium.\x0d",

                  shipid, shipid, ShipName(shipid, NULL), sourceid,
                  ShipLocationX(shipid), ShipLocationY(shipid),
                  data->gates[targetindex].id,
                  data->gates[targetindex].x, data->gates[targetindex].y,
                  config->usefuelburn);
                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                PutShipLocationX(shipid, data->gates[targetindex].x);
                PutShipLocationY(shipid, data->gates[targetindex].y);
                PutShipSpeed(shipid, 0);
                PutShipRelWaypointX(shipid, 0);
                PutShipRelWaypointY(shipid, 0);
                PutShipCargo(shipid, NEUTRONIUM, ShipCargo(shipid, NEUTRONIUM) -
                  config->usefuelburn);


                free(message);
        }
        return 0;
}

/*
 * low_addgate2list()
 * Adds one free slot to a table of stargates
 * allocates new memory for new table and
 * frees old table.
 * Proper use: gates = low_addgate2list(numgates, gates); numgates++;
 */

struct s_stargate *low_addgate2list(int listlength, struct s_stargate *gates)
{
        struct s_stargate *newgates = NULL;

        if ((newgates = malloc((listlength + 1) *
          sizeof (struct s_stargate))) == NULL) {
                fprintf(stderr,
                  "Memory allocation failure at low_addgate2list():\n%s",
                  strerror(errno));
        }

        memcpy(newgates, gates, listlength * sizeof(struct s_stargate));

        free(gates);

        return newgates;
}

int checkshipcargo(int shipid, struct s_stargatedata *data,
                   struct s_stargate_config *config)
{
        int i, j;
        char *message = NULL;


        if (IsShipExist(shipid)) {
                if ((message = malloc(600)) == NULL) {
                        fprintf(stderr, "Memory allocation failure @ checkshipcargo():\n  %s\n", strerror(errno));
                        exit(EXIT_FAILURE);
                }

                for ( i = 0 ; i < data->numships ; i++ ) {
                        if (data->ships[i] != shipid)
                                continue;
                        for ( j = 0 ; j < NumCargoTypes ; j++ ) {
                                if (ShipCargo(shipid, j) < config->buildcost2[j]) {
                                        data->ships[i] = 0;
                                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We have recycled our stargate.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                                }
                        }
                }

                free(message);
        }
        return 0;
}




int deploystargate(int shipid, struct s_stargatedata *data,
                    struct s_stargate_config *config)
{
        int i;
        char *message = NULL;
        int shipindex;
        int distfromplanet = 100000;


        if (IsShipExist(shipid)) {
                if ((message = malloc(600)) == NULL) {
                        fprintf(stderr, "Cannot malloc in deploystargate(): %s\n",
                          strerror(errno));
                        exit(EXIT_FAILURE);
                }

                PutShipMission(shipid, 5);

                if (config->allow_deploysg_p[ShipOwner(shipid)] == 0 ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Game configuration says thou\x0d"
"may not deploy stargates,\x0d"
"master.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }

                if (config->allow_deploysg_h[ShipHull(shipid)] == 0 ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Game configuration says this\x0d"
"type of ship can not deploy\x0d"
"stargates, master.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }



                for ( i = 0 ; i < data->numships ; i++ )
                        if (data->ships[i] == shipid)
                                goto shiphasgate;

                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We are not carrying a stargate.\x0d",
                  shipid, shipid, ShipName(shipid, NULL));
                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                free(message);
                return 1;


        shiphasgate:
                shipindex = i;

                if (data->gates != NULL)
                        for ( i = 0 ; i < data->numgates ; i++ ) {
                                if (ShipLocationX(shipid) == data->gates[i].x &&
                                    ShipLocationY(shipid) == data->gates[i].y ) {
                                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"There is already a stargate at\x0d"
"our current position.\x0d",
                                          shipid, shipid, ShipName(shipid, NULL));
                                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                                        free(message);
                                        return 1;
                                }
                        }


                for ( i = 1 ; i <= PLANET_NR ; i++ ) {
                        if (! IsPlanetExist(i) )
                                continue;
                        distfromplanet = min(distfromplanet, Distance(
                          ShipLocationX(shipid), ShipLocationY(shipid),
                          PlanetLocationX(i), PlanetLocationY(i)));
                }

                if ( distfromplanet < config->mindistfromplanet ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We are too close to a planet\x0d"
"to deploy our stargate.\x0d"
"Minimum distance is %u lys.\x0d",
                          shipid, shipid, ShipName(shipid, NULL),
                          config->mindistfromplanet);
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }


                if (ShipCargo(shipid, NEUTRONIUM) == 0) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We do not have enough\x0d"
"fuel to deploy our stargate.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }


/* Ok, deploy the gate */

/* Remove minerals */
                for ( i = 0 ; i < NumCargoTypes ; i++ )
                        PutShipCargo(shipid, i,
                          ShipCargo(shipid, i) - config->buildcost2[i]);

/* Remove the gate from the ship */

                data->ships[shipindex] = 0;


                data->gates = low_addgate2list(data->numgates, data->gates);

                data->gates[data->numgates].id = data->lastid;
                data->lastid++;
                data->gates[data->numgates].x = ShipLocationX(shipid);
                data->gates[data->numgates].y = ShipLocationY(shipid);
                data->gates[data->numgates].builtby = ShipOwner(shipid);
                data->gates[data->numgates].knownby.w = 0x01 << ShipOwner(shipid);
                data->numgates++;

                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We have deployed our stargate\x0d"
"at (%u, %u).\x0d",
                  shipid, shipid, ShipName(shipid, NULL),
                  ShipLocationX(shipid), ShipLocationY(shipid));
                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                free(message);
        }
        return 0;
}

/*
 * low_addship2list()
 * Adds one free slot to a table of ships (int)
 * allocates new memory for new table and
 * frees old table.
 * Proper use: ships = low_addship2list(numships, ships); numships++;
 */

Uns16 *low_addship2list(int listlength, Uns16 *ships)
{
        Uns16 *newships = NULL;

        if ((newships = malloc((listlength + 1) *
          sizeof (Uns16))) == NULL) {
                fprintf(stderr,
                  "Memory allocation failure at low_addship2list():\n  %s\n",
                  strerror(errno));
        }

        memcpy(newships, ships, listlength * sizeof(Uns16));
        free(ships);

        return newships;
}



int buildstargate(int shipid, struct s_stargatedata *data,
                    struct s_stargate_config *config)
{
        int i;
        int planet;
        char *message = NULL;


        if (IsShipExist(shipid)) {
                if ((message = malloc(600)) == NULL) {
                        fprintf(stderr, "malloc failed: %s\n", strerror(errno));
                        exit(EXIT_FAILURE);
                }

                PutShipMission(shipid, 5);
                PutShipSpeed(shipid, 0);
                PutShipRelWaypointX(shipid, 0);
                PutShipRelWaypointY(shipid, 0);

                if (config->allow_buildsg_p[ShipOwner(shipid)] == 0 ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Game configuration says thou\x0d"
"may not build stargates,\x0d"
"master.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }

                if (config->allow_buildsg_h[ShipHull(shipid)] == 0 ) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"Game configuration says this\x0d"
"type of ship can not build\x0d"
"stargates, master.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }


                for ( i = 0 ; i < data->numships ; i++ ) {
                        if (data->ships[i] == shipid) {
                                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We can not build a stargate\x0d"
"because we are already\x0d"
"carrying one.\x0d",
                                  shipid, shipid, ShipName(shipid, NULL));
                                WriteAUXHOSTMessage(ShipOwner(shipid), message);
                                free(message);
                                return 1;
                        }
                }


                if (ShipCargo(shipid, NEUTRONIUM) < config->buildcost1[NEUTRONIUM] + config->buildcost2[NEUTRONIUM] ||
                    ShipCargo(shipid, TRITANIUM) < config->buildcost1[TRITANIUM] + config->buildcost2[TRITANIUM] ||
                    ShipCargo(shipid, DURANIUM) < config->buildcost1[DURANIUM] + config->buildcost2[DURANIUM] ||
                    ShipCargo(shipid, MOLYBDENUM) < config->buildcost1[MOLYBDENUM] + config->buildcost2[MOLYBDENUM] ||
                    ShipCargo(shipid, COLONISTS) < config->buildcost1[COLONISTS] + config->buildcost2[COLONISTS] ||
                    ShipCargo(shipid, SUPPLIES) < config->buildcost1[SUPPLIES] + config->buildcost2[SUPPLIES] ||
                    ShipCargo(shipid, CREDITS) < config->buildcost1[CREDITS] + config->buildcost2[CREDITS]) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We can not build a stargate\x0d"
"because we do not have enough\x0d"
"minerals.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }




                for ( i = 1 ; i <= PLANET_NR ; i++ ) {
                        if (IsPlanetExist(i) &&
                            PlanetLocationX(i) == ShipLocationX(shipid) &&
                            PlanetLocationY(i) == ShipLocationY(shipid) &&
                            PlanetOwner(i) == ShipOwner(shipid) &&
                            IsBaseExist(i)) {
                                goto planetfound;
                        }
                }

                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We cannot build a stargate\x0d"
"because we do not own a base\x0d"
"at this position.\x0d",
                  shipid, shipid, ShipName(shipid, NULL));
                  WriteAUXHOSTMessage(ShipOwner(shipid), message);
                free(message);
                return 1;


        planetfound:
                planet = i;




                if (BaseTech(planet, HULL_TECH) < config->minhulltechtobuild) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We cannot build a stargate\x0d"
"because our base's hull\x0d"
"tech level is too low.\x0d"
"%u is required and ours is %u\x0d",
                          shipid, shipid, ShipName(shipid, NULL),
                          config->minhulltechtobuild, BaseTech(planet, HULL_TECH));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }

                if (ShipCargo(shipid, NEUTRONIUM) == 0) {
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We cannot build a stargate\x0d"
"because we do not have enough\x0d"
"fuel.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }

                for ( i = 1 ; i <= SHIP_NR ; i++ ) {
                        if (! IsShipExist(i))
                                continue;
                        if (ShipLocationX(i) != ShipLocationX(shipid) ||
                            ShipLocationY(i) != ShipLocationY(shipid))
                                continue;
                        if (ShipMission(i) != 7 || ShipTowTarget(i) != shipid)
                                continue;
                        if (ShipCargo(i, NEUTRONIUM) == 0 ||
                            (HullEngineNumber(ShipHull(i)) < 2 &&
                             gPconfigInfo->OneTow))
                                continue;
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We cannot build a stargate\x0d"
"because we are under tow!\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                        free(message);
                        return 1;
                }




/* They can build a stargate, so let's do it. */

/* Remove minerals */
                for ( i = 0 ; i < NumCargoTypes ; i++ )
                        PutShipCargo(shipid, i,
                          ShipCargo(shipid, i) - config->buildcost1[i]);



                data->ships = low_addship2list(data->numships, data->ships);
                data->ships[data->numships] = shipid;
                data->numships++;

                sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We have built a stargate.\x0d",
                          shipid, shipid, ShipName(shipid, NULL));
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);

                free(message);
        }
        return 0;
}

int alliedvision(struct s_stargatedata *data,
                 struct s_stargate_config *config)
{
        int i, j, k;

        for ( i = 1 ; i <= RACE_NR ; i++ ) {
                for ( j = 1 ; j <= RACE_NR ; j++ ) {
                        if ( i == j)
                                continue;
                        if ( ! (PlayersAreAllies(i, j) &&
                                PlayerAllowsAlly(i, j, ALLY_SHIPS)))
                                continue;
                        for ( k = 0 ; k < data->numgates ; k++ ) {
                                if (data->gates[k].knownby.w & (0x01 << i))
                                        data->gates[k].knownby.w |= (0x01 << j);
                        }
                }
        }
        return 0;
}


int scanstargates(int shipid, struct s_stargatedata *data,
                    struct s_stargate_config *config)
{
        int i, j;
        double chance;
        int distance;
        char *message = NULL;
        int gateisonplanet;


        if (IsShipExist(shipid)) {
                if ((message = malloc(600)) == NULL) {
                        fprintf(stderr, "Memory allocation failure at scanstargates():\n  %s\n", strerror(errno));
                        exit(EXIT_FAILURE);
                }

                if (config->verbosity >= 3)
                        fprintf(logfile, " Ship %d scans for stargates... \n", shipid);

                if (ShipCargo(shipid, NEUTRONIUM) == 0) {
                        if (config->verbosity >= 3)
                                fprintf(logfile, "  but doesn't have any fuel on board.\n");
                        free(message);
                        return 0;
                }


                for ( i = 0 ; i < data->numgates ; i++ ) {
                        if (data->gates[i].x == 0 || data->gates[i].y == 0)
                                /* The gate is destroyed */
                                continue;
                        if (data->gates[i].knownby.w & (0x01 << ShipOwner(shipid)))
                                /* We already know the gate, so no need to scan it */
                                continue;

                        if (config->verbosity >= 3)
                                fprintf(logfile, "  stargate %d ", data->gates[i].id);

                        distance = Distance(ShipLocationX(shipid), ShipLocationY(shipid),
                                            data->gates[i].x, data->gates[i].y);

                        if (ShipMission(shipid) != SensorSweep && distance > 0) {
                                if (config->verbosity >= 3)
                                        fprintf(logfile, "isn't under the blind ship\n");
                                continue;
                        }

                        if (distance >= config->maxscandist) {
                                if (config->verbosity >= 3)
                                        fprintf(logfile, "is too far\n");
                                continue;
                        }

                        chance = (double) (config->maxscandist - distance) /
                                  config->maxscandist;
                        if (distance <= config->maxseedist) {
                                if (config->verbosity >= 3)
                                        fprintf(logfile, "is close enough to see\n");
                                chance = 1;
                        }

                        gateisonplanet = 0;
                        for ( j = 1 ; j <= PLANET_NR ; j++ ) {
                                if (!IsPlanetExist(j))
                                        continue;
                                if (PlanetLocationX(j) == data->gates[i].x &&
                                    PlanetLocationY(j) == data->gates[i].y) {
                                        gateisonplanet = 1;
                                        break;
                                }
                        }

                        if (gateisonplanet && distance != 0) {
                                if (config->verbosity >= 3)
                                        fprintf(logfile, "is on planet and the ship is not\n");
                                continue;
                        }

                        if ((double) (((double) rand()) / ((double) RAND_MAX)) > chance) {
                                if (config->verbosity >= 3)
                                        fprintf(logfile, "fools the scanners\n");
                                /* We didn't see the gate */
                                /* Next! */
                                continue;
                        }

                        /* We saw the gate */

                        if (config->verbosity >= 3)
                                fprintf(logfile, "gets caught\n");
                        data->gates[i].knownby.w |= (0x01 << ShipOwner(shipid));
                        sprintf(message,
"(-s%04u)<<< Stargate Message >>>\x0d\x0d"    /* sender's id */
"FROM: ID#%3u, %s\x0d"                        /* sender's id and name */
"We have detected a stargate!\x0d"
"At (%u, %u)\x0d"
"ID#%u distance %u lys\x0d"
"Owner: %u\x0d",
                          shipid, shipid, ShipName(shipid, NULL),
                          data->gates[i].x, data->gates[i].y,
                          data->gates[i].id, distance, data->gates[i].builtby);
                        WriteAUXHOSTMessage(ShipOwner(shipid), message);
                }

                free(message);
        }
        return 0;
}


int sendgatemessages(struct s_stargatedata *data,
                    struct s_stargate_config *config)
{
        int i, j;
        char *message = NULL;

        if ((message = malloc(600)) == NULL) {
                fprintf(stderr, "Memory allocation failure @ sendgatemessages():\n  %s\n", strerror(errno));
                exit(EXIT_FAILURE);
        }

        for ( i = 1 ; i <= RACE_NR ; i++ ) {
                for ( j = 0 ; j < data->numgates ; j++ ) {
                        if (data->gates[j].x == 0)
                                continue;
                        if (data->gates[j].knownby.w & (0x01 << i)) {
                                sprintf(message,
"(-s0000)<<< Known Stargate >>>\x0d\x0d"
"Stargate ID#%u at (%u, %u)\x0d"
"   Owner: %u\x0d",
                                  data->gates[j].id,
                                  data->gates[j].x,
                                  data->gates[j].y,
                                  data->gates[j].builtby);
                                WriteAUXHOSTMessage(i, message);
                        }
                }
                for ( j = 0 ; j < data->numships ; j++ ) {
                        if ((! IsShipExist(data->ships[j])) || data->ships[j] == 0)
                                continue;
                        if (ShipOwner(data->ships[j]) == i) {
                                sprintf(message,
"(-s%04u)<<< Known Stargate >>>\x0d\x0d"
"Our ship ID#%u at\x0d"
"(%u, %u) is carrying\x0d"
"a stargate\x0d",
                                  data->ships[j], data->ships[j],
                                  ShipLocationX(data->ships[j]),
                                  ShipLocationY(data->ships[j]));
                                WriteAUXHOSTMessage(i, message);

                        }
                }
        }

        free(message);
        return 0;
}


int sendconfigmessages(struct s_stargate_config *config)
{
        int i;
        int towhom[RACE_NR + 1];
        char *message = NULL;
        char fc[5];

        if ((message = malloc(600)) == NULL) {
                fprintf(stderr, "Memory allocation failure at sendconfigmessages():\n  %s\n", strerror(errno));
                exit(EXIT_FAILURE);
        }

        for ( i = 1 ; i <= RACE_NR ; i++ )
                towhom[i] = 0;

        for ( i = 1 ; i <= PLANET_NR ; i++ ) {
                if (! IsPlanetExist(i))
                        continue;
                if (PlanetOwner(i) == 0)
                        continue;
                if (strcmp(PlanetFCode(i, fc), "con") == 0)
                        towhom[PlanetOwner(i)] = 1;
        }

        for ( i = 1 ; i <= RACE_NR ; i++ ) {
                if (towhom[i] == 0)
                        continue;
                sprintf(message,
"(-s0000)<<< Stargate Config >>>\x0d\x0d"
"Stargate addon configuration:\x0d"
"Build costs of a stargate:\x0d"
"  %4u + %4u = %4u neutronium\x0d"
"  %4u + %4u = %4u tritanium\x0d"
"  %4u + %4u = %4u duranium\x0d"
"  %4u + %4u = %4u molybdenum\x0d"
"  %4u + %4u = %4u supplies\x0d"
"  %4u + %4u = %4u colonists\x0d"
"  %4u + %4u = %4u money\x0d"
"Min distance from planets %u\x0d"
"Min Hull tech to build: %u\x0d",
                  config->buildcost1[NEUTRONIUM], config->buildcost2[NEUTRONIUM], config->buildcost1[NEUTRONIUM] + config->buildcost2[NEUTRONIUM],
                  config->buildcost1[TRITANIUM], config->buildcost2[TRITANIUM], config->buildcost1[TRITANIUM] + config->buildcost2[TRITANIUM],
                  config->buildcost1[DURANIUM], config->buildcost2[DURANIUM], config->buildcost1[DURANIUM] + config->buildcost2[DURANIUM],
                  config->buildcost1[MOLYBDENUM], config->buildcost2[MOLYBDENUM], config->buildcost1[MOLYBDENUM] + config->buildcost2[MOLYBDENUM],
                  config->buildcost1[SUPPLIES], config->buildcost2[SUPPLIES], config->buildcost1[SUPPLIES] + config->buildcost2[SUPPLIES],
                  config->buildcost1[COLONISTS], config->buildcost2[COLONISTS], config->buildcost1[COLONISTS] + config->buildcost2[COLONISTS],
                  config->buildcost1[CREDITS], config->buildcost2[CREDITS], config->buildcost1[CREDITS] + config->buildcost2[CREDITS],
                  config->mindistfromplanet, config->minhulltechtobuild);
                WriteAUXHOSTMessage(i, message);

                sprintf(message,
"(-s0000)<<< Stargate Config >>>\x0d\x0d"
"Stargate addon configuration:\x0d"
"Min travel distance: %u\x0d"
"Max travel distance: %u\x0d"
"Using burns fuel: %u units\x0d"
"Ships get lost in hyperspace: %c\x0d"
"Min Engine tech to use: %u\x0d"
"Mission numbers:\x0d"
"  Build:   %3u  Deploy: %3u\x0d"
"  Destroy: %3u  Use:    %3u\x0d"
"Maximum scanning distance: %u\x0d"
"Maximum seeing distance: %u\x0d"
"Min Beam tech to destroy: %u\x0d"
"Alliance support: %c\x0d"
"Tell support: %c\x0d"
"Towing support: %c\x0d"
"Allow warp 0 tow: %c\x0d"
"Restrict interracial travelling: %c\x0d",

                  config->mintravel, config->maxtravel,
                  config->usefuelburn, (config->lost_in_hyperspace) ? 'y' : 'n',
                  config->minenginetechtouse, config->misn_buildsg,
                  config->misn_deploysg, config->misn_destroysg,
                  config->misn_usesg, config->maxscandist,
                  config->maxseedist, config->minbeamtechtodestroy,
                  config->alliancesupport ? 'y' : 'n',
                  config->tellsupport ? 'y' : 'n',
                  config->towsupport ? 'y' : 'n',
                  config->allowwarpzerotow ? 'y' : 'n',
                  config->restrict_interracial_travelling ? 'y' : 'n');
                WriteAUXHOSTMessage(i, message);
        }

        free(message);
        return 0;
}




int rtrim(char *str, const char *chars)
{
        int i, j;




        for ( i = strlen(str) - 1; ; i-- ) {
                for ( j = 0 ; chars[j] != '\0' ; j++ )
                        if (str[i] == chars[j])
                                goto theresmore;
                str[i + 1] = '\0';
                break;
            theresmore:
                continue;
        }

        return i + 1;
}





int read_config(const char *configfilename, struct s_stargate_config *config)
{
        FILE *f;
        char buf[128];
        char *ptr[2];
        int i;

        if ((f = fopen(configfilename, "r")) == NULL) {
                fprintf(stderr, "can not open %s: %s\n",
                 configfilename,  strerror(errno));
                exit(EXIT_FAILURE);
        }
        memset(config, 0, sizeof(struct s_stargate_config));

        do {
                fgets(buf, 127, f);
                if (feof(f)) {
                        break;
                }
                if (buf[0] == '#' || buf[0] == '\n') {
                        continue;
                }
                ptr[0] = strtok(buf, "=");
                ptr[1] = strtok(NULL, "=");
                i = atoi(ptr[1]);
                rtrim(ptr[0], " \t");
                if (strcmp(ptr[0], "deny_player_build_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= RACE_NR ; i++ ) {
                                        config->allow_buildsg_p[i] = 0;
                                }
                        } else {
                                config->allow_buildsg_p[i] = 0;
                        }
                } else if (strcmp(ptr[0], "allow_player_build_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= RACE_NR ; i++ ) {
                                        config->allow_buildsg_p[i] = 1;
                                }
                        } else {
                                config->allow_buildsg_p[i] = 1;
                        }
                } else if (strcmp(ptr[0], "deny_hull_build_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= HULL_NR ; i++ ) {
                                        config->allow_buildsg_h[i] = 0;
                                }
                        } else {
                                config->allow_buildsg_h[i] = 0;
                        }
                } else if (strcmp(ptr[0], "allow_hull_build_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= HULL_NR ; i++ ) {
                                        config->allow_buildsg_h[i] = 1;
                                }
                        } else {
                                config->allow_buildsg_h[i] = 1;
                        }
                }

                if (strcmp(ptr[0], "deny_player_deploy_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= RACE_NR ; i++ ) {
                                        config->allow_deploysg_p[i] = 0;
                                }
                        } else {
                                config->allow_deploysg_p[i] = 0;
                        }
                } else if (strcmp(ptr[0], "allow_player_deploy_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= RACE_NR ; i++ ) {
                                        config->allow_deploysg_p[i] = 1;
                                }
                        } else {
                                config->allow_deploysg_p[i] = 1;
                        }

                } else if (strcmp(ptr[0], "deny_hull_deploy_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= HULL_NR ; i++ ) {
                                        config->allow_deploysg_h[i] = 0;
                                }
                        } else {
                                config->allow_deploysg_h[i] = 0;
                        }
                } else if (strcmp(ptr[0], "allow_hull_deploy_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= HULL_NR ; i++ ) {
                                        config->allow_deploysg_h[i] = 1;
                                }
                        } else {
                                config->allow_deploysg_h[i] = 1;
                        }
                }

                if (strcmp(ptr[0], "deny_player_use_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= RACE_NR ; i++ ) {
                                        config->allow_usesg_p[i] = 0;
                                }
                        } else {
                                config->allow_usesg_p[i] = 0;
                        }
                } else if (strcmp(ptr[0], "allow_player_use_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= RACE_NR ; i++ ) {
                                        config->allow_usesg_p[i] = 1;
                                }
                        } else {
                                config->allow_usesg_p[i] = 1;
                        }

                } else if (strcmp(ptr[0], "deny_hull_use_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= HULL_NR ; i++ ) {
                                        config->allow_usesg_h[i] = 0;
                                }
                        } else {
                                config->allow_usesg_h[i] = 0;
                        }
                } else if (strcmp(ptr[0], "allow_hull_use_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= HULL_NR ; i++ ) {
                                        config->allow_usesg_h[i] = 1;
                                }
                        } else {
                                config->allow_usesg_h[i] = 1;
                        }
                }

                if (strcmp(ptr[0], "deny_player_destroy_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= RACE_NR ; i++ ) {
                                        config->allow_destroysg_p[i] = 0;
                                }
                        } else {
                                config->allow_destroysg_p[i] = 0;
                        }
                } else if (strcmp(ptr[0], "allow_player_destroy_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= RACE_NR ; i++ ) {
                                        config->allow_destroysg_p[i] = 1;
                                }
                        } else {
                                config->allow_destroysg_p[i] = 1;
                        }

                } else if (strcmp(ptr[0], "deny_hull_destroy_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= HULL_NR ; i++ ) {
                                        config->allow_destroysg_h[i] = 0;
                                }
                        } else {
                                config->allow_destroysg_h[i] = 0;
                        }
                } else if (strcmp(ptr[0], "allow_hull_destroy_stargate") == 0) {
                        if (i == 0) {
                                for ( i = 1 ; i <= HULL_NR ; i++ ) {
                                        config->allow_destroysg_h[i] = 1;
                                }
                        } else {
                                config->allow_destroysg_h[i] = 1;
                        }

                } else if (strcmp(ptr[0], "build_stargate_mission") == 0) {
                        config->misn_buildsg = i;
                } else if (strcmp(ptr[0], "deploy_stargate_mission") == 0) {
                        config->misn_deploysg = i;
                } else if (strcmp(ptr[0], "use_stargate_mission") == 0) {
                        config->misn_usesg = i;
                } else if (strcmp(ptr[0], "destroy_stargate_mission") == 0) {
                        config->misn_destroysg = i;
                } else if (strcmp(ptr[0], "min_distance_from_planet") == 0) {
                        config->mindistfromplanet = i;
                } else if (strcmp(ptr[0], "min_hull_tech_to_build") == 0) {
                        config->minhulltechtobuild = i;
                } else if (strcmp(ptr[0], "min_engine_tech_to_use") == 0) {
                        config->minenginetechtouse = i;
                } else if (strcmp(ptr[0], "min_beam_tech_to_destroy") == 0) {
                        config->minbeamtechtodestroy = i;
                } else if (strcmp(ptr[0], "destroy_targetless_ships") == 0 ||
                           strcmp(ptr[0], "lost_in_hyperspace") == 0) {
                        config->lost_in_hyperspace = i;
                } else if (strcmp(ptr[0], "maxscandistance") == 0) {
                        config->maxscandist = i;
                } else if (strcmp(ptr[0], "maxseedistance") == 0) {
                        config->maxseedist = i;
                } else if (strcmp(ptr[0], "max_travel_distance") == 0) {
                        config->maxtravel = i;
                } else if (strcmp(ptr[0], "min_travel_distance") == 0) {
                        config->mintravel = i;
                } else if (strcmp(ptr[0], "use_fuel_burn") == 0) {
                        config->usefuelburn = i;
                } else if (strcmp(ptr[0], "verbosity") == 0) {
                        config->verbosity = i;
                } else if (strcmp(ptr[0], "buildcost1") == 0) {
                        config->buildcost1[0] = atoi(strtok(ptr[1], ","));
                        for ( i = 1 ; i < NumCargoTypes ; i++ ) {
                                config->buildcost1[i] = atoi(strtok(NULL, ","));
                        }
                } else if (strcmp(ptr[0], "buildcost2") == 0) {
                        config->buildcost2[0] = atoi(strtok(ptr[1], ","));
                        for ( i = 1 ; i < NumCargoTypes ; i++ ) {
                                config->buildcost2[i] = atoi(strtok(NULL, ","));
                        }
                } else if (strcmp(ptr[0], "alliance_support") == 0) {
                        config->alliancesupport = i;
                } else if (strcmp(ptr[0], "tow_support") == 0) {
                        config->towsupport = i;
                } else if (strcmp(ptr[0], "allow_warp0_tow") == 0) {
                        config->allowwarpzerotow = i;
                } else if (strcmp(ptr[0], "tell_support") == 0) {
                        config->tellsupport = i;
                } else if (strcmp(ptr[0], "restrict_interracial_travelling") == 0) {
                        config->restrict_interracial_travelling = i;
                }
        } while(1);

#ifdef DEBUGSTARGATES
        printf("buildcost1 =");
        for ( i = 0 ; i < NumCargoTypes - 1; i++ )
                printf(" %u,", config->buildcost1[i]);
        printf(" %u\n", buildcost1[i]);
        printf("buildcost1 =");
        for ( i = 0 ; i < NumCargoTypes - 1; i++ )
                printf(" %u,", config->buildcost1[i]);
        printf(" %u\n", buildcost1[i]);
        printf("mindistfromplanet = %d\n", config->mindistfromplanet);
        printf("minhulltech = %d\n", config->minhulltechtobuild);
        printf("minenginetech = %d\n", config->minenginetechtouse);
        printf("minbeamtech = %d\n", config->minbeamtechtodestroy);
        printf("useburnfuel = %d\n", config->usefuelburn);
        printf("destroy = %d\n", config->destroy);
        printf("maxscandist = %d\n", config->maxscandist);
        printf("verbosity = %d\n", config->verbosity);
        printf("misn_buildsg = %d\n", config->misn_buildsg);
        printf("misn_deploysg = %d\n", config->misn_deploysg);
        printf("misn_usesg = %d\n", config->misn_usesg);
        printf("misn_destroysg = %d\n", config->misn_destroysg);
#endif

        return 0;
}

int low_cleangatelist(int num, struct s_stargate *gates)
{
        int i;
        struct s_stargate *newgates = NULL;
        int index = 0;

        if ((newgates = malloc(sizeof(struct s_stargate) * num)) == NULL) {
                fprintf(stderr, "Memory allocation failure @ low_cleangatelist():\n  %s\n", strerror(errno));
                exit(EXIT_FAILURE);
        }

        for ( i = 0 ; i < num ; i++ ) {
                if (gates[i].x != 0 &&
                    gates[i].y != 0) {
                        memcpy(&newgates[index], &gates[i], sizeof(struct s_stargate));
                        index++;
                }
        }

        memset(gates, 0, num * sizeof(struct s_stargate));

        memcpy(gates, newgates, index * sizeof(struct s_stargate));

        free(newgates);
        return index;
}


int low_cleanshiplist(int num, Uns16 *ships)
{
        int i;
        Uns16 *newships = NULL;
        int index = 0;

        if ((newships = malloc(sizeof(Uns16) * num)) == NULL) {
                fprintf(stderr, "Memory allocation failure @ low_cleanshiplist():\n  %s\n", strerror(errno));
                exit(EXIT_FAILURE);
        }

        for ( i = 0 ; i < num ; i++ ) {
                if (ships[i] != 0) {
                        newships[index] = ships[i];
                        index++;
                }
        }

        memset(ships, 0, num * sizeof(Uns16));

        memcpy(ships, newships, index * sizeof(Uns16));

        free(newships);
        return index;
}

int write_sgdata(struct s_stargatedata *data)
{
        FILE *f;
        char filename[128];
        struct s_stargatehst_header header;
#ifdef MSDOS
        sprintf(filename, "%s\\stargate.hst", gGameDirectory);
#else
        sprintf(filename, "%s/stargate.hst", gGameDirectory);
#endif
        if ((f = fopen(filename, "wb+")) == NULL) {
                perror("fopen: stargate.hst");
                exit(EXIT_FAILURE);
        }

        data->numgates = low_cleangatelist(data->numgates, data->gates);
        data->numships = low_cleanshiplist(data->numships, data->ships);


        memcpy(header.stargates, "stargates\0", 10);
        header.version = program_version;
        header.headerlen = sizeof (struct s_stargatehst_header);
        header.footerlen = 0;
        header.lastid = data->lastid;
        header.numgates = data->numgates;
        header.numships = data->numships;
        header.ptr2gates = header.headerlen;
        header.ptr2ships = header.headerlen +
                           header.numgates * sizeof (struct s_stargate);

        fwrite(&header, sizeof(struct s_stargatehst_header), 1, f);
        fwrite(data->gates, sizeof(struct s_stargate),
               data->numgates, f);
        fwrite(data->ships, sizeof(Uns16), data->numships, f);

        fclose(f);

        return 0;
}




int read_sgdata(struct s_stargatedata *stargatedata)
{
        FILE *f;
        char filename[128];
        struct s_stargatehst_header header;

        memcpy(header.stargates, "stargates\0", 10);
        header.version = program_version;
        header.headerlen = sizeof (struct s_stargatehst_header);
        header.footerlen = 0;
        header.lastid = 0;
        header.numgates = 0;
        header.numships = 0;
        header.ptr2gates = 0;
        header.ptr2ships = 0;

#ifdef MSDOS
        sprintf(filename, "%s\\stargate.hst", gGameDirectory);
#else
        sprintf(filename, "%s/stargate.hst", gGameDirectory);
#endif
        if ((f = fopen(filename, "rb+")) == NULL) {
                if ((f = fopen(filename, "wb+")) == NULL) {
                        perror("fopen: stargate.hst");
                        exit(EXIT_FAILURE);
                }
                fwrite(&header, sizeof(struct s_stargatehst_header), 1, f);
                fseek(f, 0, SEEK_SET);
        }

        fread(&header, sizeof(struct s_stargatehst_header), 1, f);

        if (strcmp(header.stargates, "stargates") != 0) {
                fputs("stargate.hst fails identification!\n", stderr);
                exit(EXIT_FAILURE);
        }

        if (header.version > program_version) {
                fputs("stargate.hst has version number bigger than this program\n", stderr);
                exit(EXIT_FAILURE);
        }

        stargatedata->numgates = header.numgates;
        if (header.numgates == 0)
                stargatedata->gates = NULL;
        else if ((stargatedata->gates =
                  calloc(header.numgates, sizeof(struct s_stargate))) == NULL) {
                perror("calloc failed");
                exit(EXIT_FAILURE);
        }

        stargatedata->numships = header.numships;
        if (header.numships == 0)
                stargatedata->ships = NULL;
        else if ((stargatedata->ships =
                  calloc(header.numships, sizeof(Uns16))) == NULL) {
                perror("calloc");
                exit(EXIT_FAILURE);
        }

        stargatedata->lastid = header.lastid;

        fseek(f, header.ptr2gates, SEEK_SET);
        fread(stargatedata->gates, sizeof(struct s_stargate), header.numgates, f);

        fseek(f, header.ptr2ships, SEEK_SET);
        fread(stargatedata->ships, sizeof(Uns16), header.numships, f);

        fclose(f);

        return 0;
}



Uns16 ShipFreeCargo(Uns16 shipid)
{
        return HullCargoCapacity(ShipHull(shipid)) -
               ShipCargo(shipid, TRITANIUM) - ShipCargo(shipid, DURANIUM) -
               ShipCargo(shipid, MOLYBDENUM) - ShipCargo(shipid, SUPPLIES) -
               ShipCargo(shipid, COLONISTS) - ShipAmmunition(shipid);
}
